package acl

import (
	"bufio"
	"fmt"
	"os"
	"strings"

	"git.autistici.org/ai3/tools/replds2/common"
)

type ACLOp int

const (
	OpRead = iota
	OpWrite
)

func (op ACLOp) String() string {
	switch op {
	case OpRead:
		return "READ"
	case OpWrite:
		return "WRITE"
	}
	return "<ERROR>"
}

type NullManager struct{}

func NewNullManager() *NullManager { return new(NullManager) }

func (m *NullManager) Check(identity string, op ACLOp, path string) bool { return true }

type Entry struct {
	Identity string
	Path     string
	Op       ACLOp
}

type Manager struct {
	acls map[Entry]struct{}
}

func NewManager(acls []Entry) *Manager {
	m := &Manager{acls: make(map[Entry]struct{})}
	for _, e := range acls {
		m.acls[e] = struct{}{}
	}
	return m
}

func (m *Manager) Check(identity string, op ACLOp, path string) bool {
	e := Entry{Identity: identity, Op: op}
	for _, p := range common.PathPrefixes(path) {
		e.Path = p
		if _, ok := m.acls[e]; ok {
			return true
		}
	}
	return false
}

func Load(path string) (*Manager, error) {
	f, err := os.Open(path)
	if err != nil {
		return nil, err
	}
	defer f.Close()

	var acls []Entry
	scanner := bufio.NewScanner(f)
	lineno := 0
	for scanner.Scan() {
		lineno++
		line := strings.TrimSpace(scanner.Text())
		if strings.HasPrefix(line, "#") || line == "" {
			continue
		}
		fields := strings.Fields(line)
		if len(fields) != 3 {
			return nil, fmt.Errorf("syntax error at %s:%d: not enough fields", path, lineno)
		}
		op, err := parseOp(fields[2])
		if err != nil {
			return nil, fmt.Errorf("syntax error at %s:%d: %w", path, lineno, err)
		}
		acls = append(acls, Entry{
			Identity: fields[0],
			Path:     fields[1],
			Op:       op,
		})
	}

	if err := scanner.Err(); err != nil {
		return nil, err
	}

	return NewManager(acls), nil
}

func parseOp(s string) (op ACLOp, err error) {
	switch strings.ToLower(s) {
	case "read", "r":
		op = OpRead
	case "write", "w":
		op = OpWrite
	default:
		err = fmt.Errorf("unknown op '%s'", s)
	}
	return
}
