package mem

import (
	"io/ioutil"
	"log"
	"os"
	"path/filepath"
	"strings"

	pb "git.autistici.org/ai3/tools/replds2/proto"
	"github.com/armon/go-radix"
)

type logTree struct {
	*radix.Tree
	*appendLog
}

func (t *logTree) get(path string) (*pb.Node, bool) {
	value, ok := t.Tree.Get(path)
	if ok {
		return value.(*pb.Node), true
	}
	return nil, false
}

func (t *logTree) logPut(node *pb.Node) error {
	t.Tree.Insert(node.Path, node)
	return nil
}

func (t *logTree) put(node *pb.Node) error {
	t.Tree.Insert(node.Path, node)
	return t.appendLog.Write(node)
}

func (t *logTree) dump(f func(*pb.Node) error) (err error) {
	t.Tree.Walk(func(path string, value interface{}) bool {
		err = f(value.(*pb.Node))
		return err != nil
	})
	return
}

func newLogTree(path string) (*logTree, error) {
	tree := &logTree{Tree: radix.New()}
	log, err := openLog(path, tree.logPut, tree.dump)
	if err != nil {
		return nil, err
	}
	tree.appendLog = log
	return tree, nil
}

type MemStore struct {
	tree *logTree
}

func NewMemStore(path string) (*MemStore, error) {
	tree, err := newLogTree(path)
	if err != nil {
		return nil, err
	}
	return &MemStore{
		tree: tree,
	}, nil
}

func (m *MemStore) AddNode(node *pb.Node) (bool, error) {
	if old, ok := m.tree.get(node.Path); ok && old.Version >= node.Version {
		return false, nil
	}
	return true, m.tree.put(node)
}

func (m *MemStore) Summary() *pb.NodeStateSummary {
	summary := &pb.NodeStateSummary{
		NodeMeta: make([]*pb.NodeMeta, 0, m.tree.Len()),
	}
	m.tree.Walk(func(path string, value interface{}) bool {
		summary.NodeMeta = append(summary.NodeMeta, value.(*pb.Node).Meta())
		return false
	})
	return summary
}

func (m *MemStore) TreeDiff(summary *pb.NodeStateSummary, prefix string) []*pb.Node {
	if !strings.HasSuffix(prefix, "/") {
		prefix += "/"
	}

	// Convert the summary to a map, so we can look it up quickly.
	tmp := make(map[string]int64)
	for _, nm := range summary.NodeMeta {
		tmp[nm.Path] = nm.Version
	}

	// Now scan our tree and accumulate missing elements.
	var out []*pb.Node
	m.tree.WalkPrefix(prefix, func(path string, value interface{}) bool {
		node := value.(*pb.Node)
		version, ok := tmp[path]
		if !ok || version < node.Version {
			out = append(out, node)
		}
		return false
	})
	return out
}

func (m *MemStore) Close() {
	m.tree.Close()
}

// MemStore, but with node data backed by the file system.
type MemFileStore struct {
	*MemStore

	path string
}

func NewMemFileStore(path string) (*MemFileStore, error) {
	ms, err := NewMemStore(filepath.Join(path, ".db"))
	if err != nil {
		return nil, err
	}
	return &MemFileStore{
		MemStore: ms,
		path:     path,
	}, nil
}

func (m *MemFileStore) nodePath(node *pb.Node) string {
	return filepath.Join(m.path, node.Path)
}

func (m *MemFileStore) AddNode(node *pb.Node) (bool, error) {
	ok, _ := m.MemStore.AddNode(node.WithoutData()) // nolint: errcheck
	if !ok {
		return false, nil
	}

	path := m.nodePath(node)
	if node.Deleted {
		err := os.Remove(path)
		if os.IsNotExist(err) {
			err = nil
		}
		return false, err
	}

	if err := os.MkdirAll(filepath.Dir(path), 0700); err != nil {
		return false, err
	}
	log.Printf("writing %s data to %s", node.Path, path)
	return true, ioutil.WriteFile(path, node.Data, 0600)
}

func (m *MemFileStore) TreeDiff(summary *pb.NodeStateSummary, prefix string) []*pb.Node {
	nodes := m.MemStore.TreeDiff(summary, prefix)
	out := make([]*pb.Node, 0, len(nodes))
	for _, node := range nodes {
		if node.Deleted {
			out = append(out, node)
			continue
		}
		data, err := ioutil.ReadFile(m.nodePath(node))
		if err != nil {
			log.Printf("inconsistency: data for node %s not found, dropping node", node.Path)
			m.MemStore.tree.Delete(m.nodePath(node))
		}
		out = append(out, node.WithData(data))
	}
	return out
}
