package watcher

import (
	"encoding/json"
	"io/ioutil"
	"log"
	"os"
	"path/filepath"
	"strings"

	"git.autistici.org/ai3/tools/replds2/common"
	pb "git.autistici.org/ai3/tools/replds2/proto"
)

type Trigger interface {
	Run([]*pb.Node) error
}

type triggerManager map[string]Trigger

func (m triggerManager) Has(path string) bool {
	_, ok := m[path]
	return ok
}

func (m triggerManager) newBatch() *common.NotifyBatch {
	return common.NewNotifyBatch(m)
}

func (m triggerManager) Notify(b *common.NotifyBatch) {
	b.Apply(func(path string, nodes []*pb.Node) {
		trigger := m[path]
		trigger.Run(nodes)
	})
}

type scriptTrigger struct {
	Path    string `json:"path"`
	Command string `json:"command"`
}

func (t *scriptTrigger) Run(nodes []*pb.Node) error {
	return nil
}

func LoadTriggersFromDir(dir string) (map[string]Trigger, error) {
	d, err := os.Open(dir)
	if err != nil {
		return nil, err
	}
	defer d.Close()

	m := make(map[string]Trigger)
	files, err := d.Readdir(0)
	if err != nil {
		return nil, err
	}
	for _, f := range files {
		if !f.Mode().IsRegular() {
			continue
		}
		if strings.Contains(f.Name(), ".") {
			continue
		}
		data, err := ioutil.ReadFile(filepath.Join(dir, f.Name()))
		if err != nil {
			log.Printf("couldn't read %s: %v", f.Name(), err)
			continue
		}
		var trig scriptTrigger
		if err := json.Unmarshal(data, &trig); err != nil {
			log.Printf("invalid JSON in %s: %v", f.Name(), err)
			continue
		}
		if trig.Path == "" || trig.Command == "" {
			log.Printf("invalid trigger specification in %s", f.Name())
			continue
		}

		m[trig.Path] = &trig
	}

	return m, nil
}
