package watcher

import (
	"context"
	"io"
	"log"
	"time"

	replds "git.autistici.org/ai3/tools/replds2"
	pb "git.autistici.org/ai3/tools/replds2/proto"
	"google.golang.org/grpc"
	"google.golang.org/grpc/codes"
	"google.golang.org/grpc/status"
)

type Watcher struct {
	path     string
	store    replds.Store
	triggers triggerManager
	conn     *grpc.ClientConn
}

func New(conn *grpc.ClientConn, store replds.Store, path string, triggers map[string]Trigger) *Watcher {
	return &Watcher{
		path:     path,
		conn:     conn,
		store:    store,
		triggers: triggerManager(triggers),
	}
}

func (w *Watcher) Run(ctx context.Context) {
	stub := pb.NewRepldsClient(w.conn)
	for {
		// Establish the streaming connection.
		req := &pb.WatchRequest{
			BasePath: w.path,
			Summary:  w.store.Summary(),
		}
		stream, err := stub.Watch(ctx, req)
		if err == context.Canceled || status.Code(err) == codes.Canceled {
			return
		}
		if err != nil {
			log.Printf("watch error: %v", err)
			goto retry
		}

		for {
			resp, err := stream.Recv()
			if err == io.EOF {
				break
			}
			if err != nil {
				log.Printf("watch stream receive error: %v", err)
				break
			}

			// Run triggers for each batch.
			tb := w.triggers.newBatch()
			for _, node := range resp.Nodes {
				w.store.AddNode(node)
				tb.Add(node)
			}
			w.triggers.Notify(tb)
		}

	retry:
		time.Sleep(1 * time.Second)
	}

}
